class BigInt:
    def __init__(self, val: int) -> None: ...

class SmallInt:
    def __init__(self, val: int) -> None: ...

class TinyInt:
    def __init__(self, val: int) -> None: ...

class Double:
    def __init__(self, val: float) -> None: ...

class Counter:
    def __init__(self, val: int) -> None: ...
